@@ -1,7 +1,8 @@
-__author__ = 'Tanapat Kahabodeekanokkul (pahntanapat@gmail.com)'
+__author__ = 'Tanapat Kahabodeekanokkul (GitHub @pahntanapat, pahntanapat@gmail.com), Natthaphong Ratchakhom (GitHub @natthaphong17)'
 
 from concurrent.futures import (
     CancelledError,
+    Executor,
     ThreadPoolExecutor,
     ProcessPoolExecutor,
     Future,
@@ -18,15 +19,114 @@ from concurrent.futures.thread import (
 )
 
 import gc
-from logging import Logger, getLogger
 from multiprocessing import Pipe, freeze_support, get_context, cpu_count
 from multiprocessing.connection import Connection
-from typing import Any, Callable, Optional, Tuple, Union
+import sys
+import time
+import traceback
+from typing import Any, Callable, Iterable, Iterator, Optional, Tuple, TypeVar, Union
 
 gc.enable()
 
+_T = TypeVar("_T")
 
-class AWSPoolFuture(Future):
+
+def function_wrapper(fn, *args, **kwargs):
+    """Wraps `fn` in order to preserve the traceback of any kind of
+    raised exception
+
+    """
+    try:
+        return fn(*args, **kwargs)
+    except Exception:
+        raise sys.exc_info()[0](traceback.format_exc())  # Creates an
+        # exception of the
+        # same type with the
+        # traceback as
+        # message
+
+
+class StackTraceExecutor(Executor):
+
+    """Base class for Executor, attaching stack trace when execption occurs.
+
+    We acknowledge [se7entyse7en](https://stackoverflow.com/users/3276106/se7entyse7en)'s [answer in StackOverflow](https://stackoverflow.com/a/24457608).
+    """
+
+    def submit(self, fn, *args, **kwargs):
+        """Submits the wrapped function instead of `fn`"""
+
+        return super().submit(function_wrapper, fn, *args, **kwargs)
+
+    def map_args(self,
+                 fn: Callable[..., _T],
+                 iterable: Iterable[Any],
+                 *args,
+                 timeout: Optional[float] = None,
+                 chunksize: Optional[int] = None) -> Iterator[_T]:
+        l = len(iterable)
+        iterables = [iterable] + [[i] * l for i in args]
+        return super().map(fn,
+                           *iterables,
+                           timeout=timeout,
+                           chunksize=chunksize)
+
+    def map(self, fn: Callable[..., _T], *iterables, timeout: Optional[float] = None, chunksize: int = 1) -> Iterator[_T]:
+        """Returns an iterator equivalent to map(fn, iter).
+
+        Args:
+            fn: A callable that will take as many arguments as there are
+                passed iterables.
+            timeout: The maximum number of seconds to wait. If None, then there
+                is no limit on the wait time.
+            chunksize: The size of the chunks the iterable will be broken into
+                before being passed to a child process. This argument is only
+                used by ProcessPoolExecutor; it is ignored by
+                ThreadPoolExecutor.
+
+        Returns:
+            An iterator equivalent to: map(func, *iterables) but the calls may
+            be evaluated out-of-order.
+
+        Raises:
+            TimeoutError: If the entire result iterator could not be generated
+                before the given timeout.
+            Exception: If fn(*args) raises for any values.
+        """
+        if timeout is not None:
+            end_time = timeout + time.monotonic()
+
+        fs = [self.submit(fn, *args) for args in zip(*iterables)]
+
+        # Yield must be hidden in closure so that the futures are submitted
+        # before the first iterator value is required.
+        def result_iterator():
+            try:
+                # reverse to keep finishing order
+                fs.reverse()
+                while fs:
+                    # Careful not to keep a reference to the popped future
+                    if timeout is None:
+                        yield fs.pop().result()
+                    else:
+                        yield fs.pop().result(end_time - time.monotonic())
+            except Exception:
+                raise sys.exc_info()[0](traceback.format_exc())
+            finally:
+                for future in fs:
+                    future.cancel()
+        return result_iterator()
+
+
+class StackTracedThreadPoolExecutor(StackTraceExecutor, ThreadPoolExecutor):
+    pass
+
+
+class StackTracedProcessPoolExecutor(StackTraceExecutor, ProcessPoolExecutor):
+    pass
+
+
+class PipePoolFuture(Future):
     def cancel(self):
         """Cancel the future if possible.
 
@@ -52,11 +152,13 @@ class _ProcessWorkItem(_WorkItem):
 
     def __init__(
         self,
-        future: AWSPoolFuture,
+        future: PipePoolFuture,
         fn: Callable,
-        logger: Logger,
+        # logger: Logger,
         args,
         kwargs,
+        initializer: Optional[Callable] = None,
+        initargs: Tuple[Any, ...] = tuple(),
         mp_method: Optional[str] = None,
         thread_sleep: Optional[float] = 0.001
     ):
@@ -64,7 +166,9 @@ class _ProcessWorkItem(_WorkItem):
         self.fn = fn
         self.args = args
         self.kwargs = kwargs
-        self.logger = logger
+        self.kwargs['initializer'] = initializer
+        self.kwargs['initargs'] = initargs
+        #self.logger = logger
         self.mp_method = mp_method
 
         if (thread_sleep) and (thread_sleep > 0):
@@ -78,7 +182,7 @@ class _ProcessWorkItem(_WorkItem):
             main, sub = Pipe()
             proc = get_context(self.mp_method).Process(
                 target=self.process_worker,
-                args=[self.fn, sub, self.logger] + list(self.args),
+                args=[self.fn, sub] + list(self.args),
                 kwargs=self.kwargs,
             )
             proc.start()
@@ -103,7 +207,7 @@ class _ProcessWorkItem(_WorkItem):
                                fn=self.fn.__name__,
                                pid=proc.pid, exitcode=proc.exitcode())
 
-                    self.logger.info(status, **kws)
+                    #self.logger.info(status, **kws)
                     raise CancelledError(status.format_map(kws))
                 raise CancelledError()
             else:
@@ -118,34 +222,46 @@ class _ProcessWorkItem(_WorkItem):
                     )
                 )
 
-        except BaseException as exc:
-            self.logger.exception('Exception from run process: {}', exc)
-            self.future.set_exception(exc)
+        except:
+            # self.logger.exception('Exception from run process: {}', exc)
+            self.future.set_exception(
+                sys.exc_info()[0](traceback.format_exc()))
             # Break a reference cycle with the exception 'exc'
             self = None
 
     @staticmethod
-    def process_worker(fn: Callable, pipe: Connection, logger: Logger, /, *arg, **kw):
+    def process_worker(fn: Callable, pipe: Connection, /, *arg, initializer: Optional[Callable] = None, initargs: Tuple[Any, ...] = tuple(), **kw):
         try:
+            if initializer is not None:
+                initializer(*initargs)
             result = fn(*arg, **kw)
             pipe.send([True, result])
-        except Exception as e:
-            logger.exception("{e} from {fn}(*{arg},**{kw})",
-                             e=e, fn=fn, arg=arg, kw=kw)
-            pipe.send([False, e])
+        except:
+            pipe.send([False, sys.exc_info()[0](traceback.format_exc())])
 
         pipe.close()
         return gc.collect()
 
 
-class AWSLambdaProcessPoolExecutor(ThreadPoolExecutor):
+class PipeProcessPoolExecutor(StackTracedThreadPoolExecutor):
+    """ProcessPoolExecutor for environment without /dev/shm
+    PipeProcessPoolExecutor enable multi-cpu parallel processing in Python as same as ProcessPoolExecutor in environment without /dev/shm (shared memory for processes) support, i.e. AWS Lambda.
+
+    Acknowlegdement
+    ---------------
+     - [Parallel Processing in Python with AWS Lambda, AWS Compute Blog](https://aws.amazon.com/th/blogs/compute/parallel-processing-in-python-with-aws-lambda/)
+     - [AWS Lambda Memory Vs CPU configuration](https://stackoverflow.com/a/66523153)
+
+
+    """
+
     def __init__(
         self,
         max_workers: Union[int, None] = None,
         thread_name_prefix: Optional[str] = None,
         initializer: Optional[Callable[..., None]] = None,
         initargs: Optional[Tuple[Any, ...]] = None,
-        logger: Optional[Logger] = None,
+        # logger: Optional[Logger] = None,
         mp_method: Optional[str] = None,
         thread_sleep: Optional[float] = None
     ) -> None:
@@ -154,7 +270,7 @@ class AWSLambdaProcessPoolExecutor(ThreadPoolExecutor):
             max_workers = 1 + (cpu_count() * 2)
 
         super().__init__(max_workers, thread_name_prefix, initializer, initargs)
-        self.logger = getLogger() if (logger is None) else logger
+        # self.logger = getLogger() if (logger is None) else logger
         self.mp_method = mp_method
         self.thread_sleep = thread_sleep
 
@@ -171,9 +287,14 @@ class AWSLambdaProcessPoolExecutor(ThreadPoolExecutor):
                     "cannot schedule new futures after interpreter shutdown"
                 )
 
-            f = AWSPoolFuture()
-            w = _ProcessWorkItem(f, fn, self.logger, args, kwargs,
-                                 mp_method=self.mp_method, thread_sleep=self.thread_sleep)
+            f = PipePoolFuture()
+            w = _ProcessWorkItem(f, fn,
+                                 # self.logger,
+                                 args, kwargs,
+                                 initializer=self._initializer,
+                                 initargs=self._initargs,
+                                 mp_method=self.mp_method,
+                                 thread_sleep=self.thread_sleep)
 
             self._work_queue.put(w)
             self._adjust_thread_count()
@@ -206,13 +327,13 @@ class Pool:
 
                 try:
                     if aws_lambda:
-                        self.pool = AWSLambdaProcessPoolExecutor(
+                        self.pool = PipeProcessPoolExecutor(
                             max_worker=pool_worker)
                     else:
-                        self.pool = ProcessPoolExecutor(
+                        self.pool = StackTracedProcessPoolExecutor(
                             max_workers=pool_worker)
                 except OSError:
-                    self.pool = AWSLambdaProcessPoolExecutor(
+                    self.pool = PipeProcessPoolExecutor(
                         max_worker=pool_worker)
             else:
                 if (pool_worker is None) or (pool_worker == 0):
@@ -220,7 +341,8 @@ class Pool:
                 elif pool_worker < 0:
                     pool_worker = min(max(1, (4 * cpu) + 1 - pool_worker), 61)
 
-                self.pool = ThreadPoolExecutor(max_workers=pool_worker)
+                self.pool = StackTracedThreadPoolExecutor(
+                    max_workers=pool_worker)
 
         else:
             self.pool = process_pool
